require "defines"

remote.add_interface("SingleSplitter",
{
	clean = function()
		global.SingleSplitter = {}
	end
})

script.on_init(function() OnInit() end)
script.on_event(defines.events.on_tick, function(event) OnTick(event) end)
script.on_event(defines.events.on_built_entity, function(event) OnBuiltEntity(event) end)
script.on_event(defines.events.on_entity_died, function(event) OnEntityDied(event) end)
script.on_event(defines.events.on_preplayer_mined_item, function(event) OnPrePlayerMinedItem(event) end)
script.on_event(defines.events.on_player_rotated_entity, function(event) OnPlayerRotatedEntity(event) end)
script.on_event(defines.events.on_robot_built_entity, function(event) OnRobotBuiltEntity(event) end)
script.on_event(defines.events.on_robot_pre_mined, function(event) OnRobotPreMined(event) end)

local north = defines.direction.north
local east = defines.direction.east
local south = defines.direction.south
local west = defines.direction.west

function OnInit()
	global.SingleSplitter = {	}
end

function OnPlayerRotatedEntity(_Event)
	if IsSingleSplitter(_Event.entity) then
		local _Splitter = _Event.entity
		for key,value in pairs(global.SingleSplitter) do 
		  if value.pos.x == _Splitter.position.x and value.pos.y == _Splitter.position.y then
			value.dir = _Splitter.direction
			value.cachedInBelt = nil
			value.cachedOutBelt = nil
			break
		  end 
		end
	end
end


local VV = 0.55
local HH = 0.23
local targetOff = {[0] = {left = {x = -HH, y = -VV}, right = {x =  HH, y = -VV}},
				   [2] = {left = {x =  VV, y = -HH}, right = {x =  VV, y =  HH}},
				   [4] = {left = {x =  HH, y =  VV}, right = {x = -HH, y =  VV}},
				   [6] = {left = {x = -VV, y =  HH}, right = {x = -VV, y = -HH}}
						 }

function MoveStack12(_ItemName, _Splitter, _SrcLine, _OutPutLine)

	local outBelt = _Splitter.cachedOutBelt
	if nil == outBelt or not outBelt.valid or not outBelt.direction == _Splitter.dir then
		_Splitter.cachedOutBelt = nil
		outBelt = nil
		
		local facing_directions = {[north] = south, [east] = west, [south] = north, [west] = east}
		local scanArea = GetScanArea(facing_directions[_Splitter.dir], _Splitter.pos)
		local belt = game.get_surface(1).find_entities_filtered{type = "transport-belt", area = scanArea}
		if belt[1] ~= nil and belt[1].direction == _Splitter.dir then
			outBelt = belt[1]
			_Splitter.cachedOutBelt = outBelt
		end
	end

	if nil ~= outBelt then
		local tgtLane = outBelt.get_transport_line(_OutPutLine)
		if tgtLane.can_insert_at_back() then
			local stack = {name = _ItemName, count = 1}
			_SrcLine.remove_item (stack)
			tgtLane.insert_at_back(stack)
			return true
		end
	end
	
	return false
end

function NegateSide(Side)
	if(Side == 1) then return 2
	elseif(Side == 2) then return 1
	else return 1 end
end

function OnTick(_Event)	
	if((game.tick % 5) == 0) then
		for _, splitter in pairs(global.SingleSplitter) do
			local inBelt = splitter.cachedInBelt
			
			if nil == inBelt or not inBelt.valid or not inBelt.direction == splitter.dir then
				splitter.cachedInBelt = nil
				inBelt = nil
				
				local scanArea = GetScanArea(splitter.dir, splitter.pos)
				local belt = game.get_surface(1).find_entities_filtered({	type="transport-belt", area = scanArea})
				if belt[1] ~= nil and belt[1].direction == splitter.dir then			
					inBelt = belt[1]
					splitter.cachedInBelt = inBelt
				end
			end
			
			if nil ~= inBelt then
				if splitter.lastInLane == nil then splitter.lastInLane = 1 end
				local takeFromLine = NegateSide(splitter.lastInLane)				
				for i=1,2,1 do					    
					local lane = inBelt.get_transport_line(takeFromLine)
					if not lane.can_insert_at(0) then
						local cnt=lane.get_item_count()
						local cmap=lane.get_contents()
						if cnt ~= 0 then		
							for res_name,res_count in pairs(cmap) do  
								splitter.lastOutLane = NegateSide(splitter.lastOutLane)
							    if MoveStack12(res_name, splitter, lane, splitter.lastOutLane) then 
									splitter.lastInLane = takeFromLine
								elseif MoveStack12(res_name, splitter, lane, NegateSide(splitter.lastOutLane)) then
									splitter.lastInLane = takeFromLine
								end									
							end
						end												
					end	
					takeFromLine = NegateSide(takeFromLine)
				end
			end
		end
	end
end
				   
function OnBuiltEntity(_Event)
	if IsSingleSplitter(_Event.created_entity) then
		AddSingleSplitter(_Event.created_entity)
	end
end

function OnRobotBuiltEntity(_Event)
	if IsSingleSplitter(_Event.created_entity) then
		AddSingleSplitter(_Event.created_entity)
	end
end

function OnRobotPreMined(_Event)
	if IsSingleSplitter(_Event.entity) then
		RemoveSingleSplitter(_Event.entity)
	end
end

function OnPrePlayerMinedItem(_Event)
	if IsSingleSplitter(_Event.entity) then
		RemoveSingleSplitter(_Event.entity)
	end
end

function OnEntityDied(_Event)
	if IsSingleSplitter(_Event.entity) then
		RemoveSingleSplitter(_Event.entity)
	end
end

function IsSingleSplitter(_Entity)
	return (_Entity.name == "Single Splitter")
end

function AddSet(_Set)
	table.insert(global.SingleSplitter, _Set)
end

function AddSingleSplitter(_Splitter)
	local _Position = _Splitter.position
	local _Direction = _Splitter.direction		

	local set = {pos = _Position, dir = _Direction, lastOutLane = 1, lastInLane = 1}		
	AddSet(set)	
end

function RemoveSingleSplitter(_Splitter)		
	for key,value in pairs(global.SingleSplitter) do 
      if value.pos.x == _Splitter.position.x and value.pos.y == _Splitter.position.y then
		table.remove(global.SingleSplitter, key)
		break
	  end 
    end
end


function GetScanArea(_Direction, _Position)
  local beltscan_coords = { -- Points to search for transport belts.
    [north] = {{_Position.x - 0.3, _Position.y - 1.1},{_Position.x + 0.3, _Position.y - 0.8}},
    [east] =  {{_Position.x + 1.1  , _Position.y - 0.3},{_Position.x + 0.8, _Position.y + 0.3}},
    [south] = {{_Position.x - 0.3, _Position.y + 1.1},{_Position.x + 0.3, _Position.y + 0.8}},
    [west] =  {{_Position.x - 1.1  , _Position.y - 0.3},{_Position.x - 0.8, _Position.y + 0.3}}
  }

    if _Direction == north then
      return beltscan_coords[south]
  	elseif _Direction == east then
  	 return beltscan_coords[west]
  	elseif _Direction == south then
  	 return beltscan_coords[north]
  	elseif _Direction == west then
  	 return beltscan_coords[east]
  	else
  	 return {{pos.x , pos.y},{pos.x , pos.y}}
  	end
end